package ru.dbotthepony.kstarbound.world.entities

import ru.dbotthepony.kstarbound.defs.DamageType
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.ChunkPos
import ru.dbotthepony.kstarbound.world.World
import ru.dbotthepony.kvector.vector.ndouble.Vector2d

/**
 * Интерфейс служит лишь для убирания жёсткой зависимости от класса Entity
 */
interface IEntity {
	/**
	 * The world this entity in, never changes.
	 */
	val world: World<*, *>

	/**
	 * The chunk this entity currently in, it is automatically updated on each change of [position]
	 */
	var chunk: Chunk<*, *>?

	/**
	 * Current entity position. If entity is logical, this can only be changed
	 * by external means, otherwise it always stands where it is.
	 *
	 * If entity is physical, then movement controller will update this on each physics step
	 * to match position of physical body.
	 *
	 * Setting this value will update [chunk] immediately, if [isSpawned] is true and [isRemoved] is false.
	 */
	var position: Vector2d

	/**
	 * This entity's angle in radians.
	 *
	 * Logical entities never rotate.
	 *
	 * Alive entities usually don't rotate.
	 *
	 * If entity is physical, this value is updated by movement controller on
	 * each physics step to match angle of physical body.
	 */
	var angle: Double

	/**
	 * This entity's movement controller. Even logical entities have one, but they have
	 * dummy movement controller, which does nothing.
	 *
	 * If entity is physical, this controller handle interaction with Box2D world, update angles and
	 * position and other stuff.
	 */
	val movement: MovementController<*>

	/**
	 * Whenever is this entity spawned in world ([spawn] called).
	 * Doesn't mean entity still exists in world, check it with [isRemoved]
	 */
	val isSpawned: Boolean

	/**
	 * Whenever is this entity was removed from world ([remove] called).
	 */
	val isRemoved: Boolean

	fun spawn()
	fun remove()
	fun think(delta: Double)
	fun onTouchSurface(velocity: Vector2d, normal: Vector2d)

	fun dealDamage(
		amount: Double,
		kind: String,
		type: DamageType,
	) {
		// Do nothing by default
	}
}

/**
 * Определяет из себя сущность в мире, которая имеет позицию, скорость и коробку столкновений
 */
abstract class Entity(override val world: World<*, *>) : IEntity {
	override var chunk: Chunk<*, *>? = null
		set(value) {
			if (!isSpawned) {
				throw IllegalStateException("Trying to set chunk this entity belong to before spawning in world")
			}

			if (isRemoved) {
				throw IllegalStateException("This entity was removed")
			}

			if (value == field) {
				return
			}

			val chunkPos = ChunkPos.fromTilePosition(position)

			if (value != null && chunkPos != value.pos) {
				throw IllegalStateException("Set proper position before setting chunk this Entity belongs to")
			}

			val oldChunk = field
			field = value

			if (oldChunk == null && value != null) {
				world.orphanedEntities.remove(this)
				value.addEntity(this)
			} else if (oldChunk != null && value == null) {
				world.orphanedEntities.add(this)
				oldChunk.removeEntity(this)
			} else if (oldChunk != null && value != null) {
				value.transferEntity(this, oldChunk)
			}
		}

	override var position = Vector2d()
		set(value) {
			if (field == value)
				return

			val old = field
			field = value

			movement.notifyPositionChanged()

			if (isSpawned && !isRemoved) {
				val oldChunkPos = ChunkPos.fromTilePosition(old)
				val newChunkPos = ChunkPos.fromTilePosition(value)

				if (oldChunkPos != newChunkPos) {
					chunk = world[newChunkPos]
				}
			}
		}

	override var angle: Double = 0.0
		set(value) {
			if (field == value)
				return

			field = value
			movement.notifyPositionChanged()
		}

	final override var isSpawned = false
		private set
	final override var isRemoved = false
		private set

	override fun spawn() {
		if (isSpawned)
			throw IllegalStateException("Already spawned")

		isSpawned = true
		world.entities.add(this)
		chunk = world[ChunkPos.fromTilePosition(position)]

		if (chunk == null) {
			world.orphanedEntities.add(this)
		}
	}

	override fun remove() {
		if (isRemoved)
			throw IllegalStateException("Already removed")

		isRemoved = true

		if (isSpawned) {
			world.entities.remove(this)
			chunk?.removeEntity(this)
		}

		movement.destroy()
	}

	/**
	 * Контроллер перемещения данной сущности
	 */
	abstract override val movement: MovementController<*>
	protected abstract fun thinkAI(delta: Double)

	/**
	 * Заставляет сущность "думать".
	 */
	final override fun think(delta: Double) {
		if (!isSpawned) {
			throw IllegalStateException("Tried to think before spawning in world")
		}

		movement.think(delta)
		thinkAI(delta)
	}

	override fun onTouchSurface(velocity: Vector2d, normal: Vector2d) {

	}
}
